/*==================================================================*
 * chat.c	- general handling of chat scripts
 *
 * Coypright (C) 1994, 1995 by O.Kirch, okir@monad.swb.de
 * For copying conditions, see the file COPYING.
 *==================================================================*/
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <fcntl.h>
#include <limits.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>

#include <tcl/tcl.h>

#include "slipsh.h"

struct sigaction	tty_sigoact;

static int  tty_write(Tcl_Interp *interp, char *cmd, int portfd, char *sendstr);
static int  tty_waitfor(int portfd, int count, int *index, char **string,
								int timeout);
static void tty_setalarm(int timeout);
static void tty_clralarm(void);
static void get_alrm(int signal);

/*==============================================================*
 * tty_send	send a string to the device.
 * Usage:
 *	tty send [-nonewline] tty string
 *==============================================================*/
int
tty_send(ClientData cd, Tcl_Interp *interp, int argc, char *argv[])
{
  char		*sp;
  int		newline = 1;
  FILE		*portfp;
  int		i, ret;

  struct option options[] = {
	{ "nonewline",	T_FALSE,	&newline		},
	{ NULL,		T_VOID,		NULL			}
  };

  /* Parse arguments */
  i = 1;
  if ((ret = tcl_getopt(interp, options, argc, argv, &i)) != TCL_OK) {
  	return TCL_ERROR;
  }
  if (i < argc-2) {
  	sprintf(interp->result, "%s: unknown argument: %s", argv[0], argv[i]);
  	return TCL_ERROR;
  }

  if (i != argc-2) {
  	sprintf(interp->result, "%s: need tty and string argument", argv[0]);
  	return TCL_ERROR;
  }
  if (Tcl_GetOpenFile(interp, argv[i++], 1, 1, &portfp) != TCL_OK) {
  	sprintf(interp->result, "%s: bad file argument", argv[0]);
  	return TCL_ERROR;
  }
  if (tty_write(interp, argv[0], fileno(portfp), argv[i]) != TCL_OK) {
  	return TCL_ERROR;
  }
  if (newline && ((sp = strrchr(argv[i], '\n')) == NULL || sp[1] != '\0')) {
  	return tty_write(interp, argv[0], fileno(portfp), "\n");
  }
  return TCL_OK;
}

/*==============================================================*
 * tty_recv	read from tty up to next newline
 * Usage:
 *	tty recv [-timeout seconds] tty variable ...
 * -timeout specifies the read timeout in seconds.
 *==============================================================*/
int
tty_recv(ClientData cd, Tcl_Interp *interp, int argc, char *argv[])
{
  long		timeout = 30;
  FILE		*portfp;
  char		buffer[1024];
  char		*variable = NULL;
  int		i, j, ret;
  int		skipblank = 0;
  int		portfd;
  char		c;

  struct option options[] = {
	{ "timeout",	T_INT,		&timeout		},
	{ NULL,		T_VOID,		NULL			}
  };

  /* Parse arguments */
  i = 1;
  if ((ret = tcl_getopt(interp, options, argc, argv, &i)) != TCL_OK) {
  	return TCL_ERROR;
  }

  if (i >= argc) {
  	sprintf(interp->result, "%s: need tty argument", argv[0]);
  	return TCL_ERROR;
  }
  if (Tcl_GetOpenFile(interp, argv[i++], 1, 1, &portfp) != TCL_OK) {
  	sprintf(interp->result, "%s: bad file argument", argv[0]);
  	return TCL_ERROR;
  }
  portfd = fileno(portfp);
  tty_setalarm(timeout);
  
  if (i < argc) {
  	variable = argv[i++];
  	skipblank = 1;
  }
  j = 0;
  while (read(portfd, &c, 1) == 1) {
  	if (i >= sizeof(buffer)-1) {
  		sprintf(interp->result, "%s: internal buffer overflow",
  					argv[0]);
  		tty_clralarm();
  		return TCL_ERROR;
  	}
  	if (c == '\n' || c == '\r') {
  		if (variable == NULL) {
  			Tcl_SetResult(interp, buffer, TCL_VOLATILE);
  			return TCL_OK;
  		}
		Tcl_SetVar(interp, variable, buffer, TCL_LEAVE_ERR_MSG);
  		tty_clralarm();
		return TCL_OK;
  	}
  	if (c == ' ' || c == '\t') {
  		if (skipblank) continue;
  		if (i < argc) {
  			Tcl_SetVar(interp, variable, buffer, TCL_LEAVE_ERR_MSG);
  			variable = argv[i++];
  			skipblank = 1;
  		}
  	} else {
  		skipblank = 1;
  	}
  	buffer[j++] = c;
  }
  tty_clralarm();
  return TCL_ERROR;
}

/*==============================================================*
 * tty_expect	wait for a set of different responses from the
 *		tty, and execute associated actions.
 * Usage:
 *	tty expect [-timeout seconds] tty
 *		string action string action ....
 * -timeout specifies the read timeout in seconds. An action is
 * either a TCL statement, or the | symbol, which can be used to
 * make one action refer to several patterns.
 *==============================================================*/
int
tty_expect(ClientData cd, Tcl_Interp *interp, int argc, char *argv[])
{
  long		timeout = 30;
  FILE		*portfp;
  int		*index;
  char		**string;
  int		i, j, ret;
  int		portfd;
  int		count;
  int		first, found;

  struct option options[] = {
	{ "timeout",	T_INT,		&timeout		},
	{ NULL,		T_VOID,		NULL			}
  };

  /* Parse arguments */
  i = 1;
  if ((ret = tcl_getopt(interp, options, argc, argv, &i)) != TCL_OK) {
  	return TCL_ERROR;
  }

  if (Tcl_GetOpenFile(interp, argv[i++], 1, 1, &portfp) != TCL_OK) {
  	sprintf(interp->result, "%s: bad file argument", argv[0]);
  	return TCL_ERROR;
  }
  if ((i % 2) != 0) {
  	sprintf(interp->result, "%s: odd number of expect/action arguments",
  				 argv[0]);
  	return TCL_ERROR;
  }
  first = i;
  count = (argc - first)/2;
  if ((index  = (int*)  calloc(count, sizeof(int)))   == NULL
   || (string = (char**)calloc(count, sizeof(char*))) == NULL) {
  	if (index) free(index);
  	sprintf(interp->result, "%s: malloc failed", argv[0]);
  	return TCL_ERROR;
  }
  found = -1;
  for (j=0; j < count; j++) {
  	string[j] = argv[first + (j<<1)];
  	if (string[j][0] == '\0' && found < 0) {
  		found = j;
  		sprintf(interp->result, "%s: empty expect string", argv[0]);
  	}
  }
  portfd = fileno(portfp);
  ret = TCL_ERROR;
  if (found < 0) {
	found = tty_waitfor(portfd, count, index, string, timeout);
	if (found >= 0) {
  		i = first + 1 + (found<<1);
  		while (found < argc && !strcmp(argv[i], "|")) i++;
  		if (i >= argc) {
  			sprintf(interp->result,
  					"%s: |'ed expect patterns but no "
  					"action", argv[0]);
  		} else {
  			ret = Tcl_Eval(interp, argv[i]);
  		}
  	} else {
  		sprintf(interp->result, "%s: timed out in expect script",
  					 argv[0]);
  	}
  }
  free(index);
  free(string);
  return ret;
}

/*==============================================================*
 * tty_chat	execute a chat script
 * Usage:
 *	tty chat [-timeout seconds]
 *		[-error errorstr] ...
 *		ttyfd send expect send expect ...
 *
 * Executes a chat script by writing to and reading from the
 * tty. The -error arguments specify a list of responses that are
 * considered errors. -timeout sets the timeout in seconds when
 * waiting for a response.
 *==============================================================*/
int
tty_chat(ClientData cd, Tcl_Interp *interp, int argc, char *argv[])
{
  opt_multi	error_opt;
  char		failmsg[1024];
  char		*waitstr[TTY_CHAT_MAXERRS+1];
  int		waitidx[TTY_CHAT_MAXERRS+1];
  int		timeout = 30;
  FILE		*portfp;
  int		portfd;
  int		count, found;
  int		i, ret;

  struct option options[] = {
	{ "error",	T_STRING|T_MULTI,&error_opt		},
	{ "timeout",	T_INT,		&timeout		},
	{ NULL,		T_VOID,		NULL			}
  };

  error_opt.array = waitstr+1;
  error_opt.count = 0;
  error_opt.max = TTY_CHAT_MAXERRS;

  /* Parse arguments */
  i = 1;
  if ((ret = tcl_getopt(interp, options, argc, argv, &i)) != TCL_OK) {
  	return TCL_ERROR;
  }
  if (i >= argc) {
  	sprintf(interp->result, "%s: need tty argument", argv[0]);
  	return TCL_ERROR;
  }
  count = error_opt.count+1;

  if (Tcl_GetOpenFile(interp, argv[i++], 1, 1, &portfp) != TCL_OK) {
  	sprintf(interp->result, "%s: bad file argument", argv[0]);
  	return TCL_ERROR;
  }
  portfd = fileno(portfp);

  while (i < argc) {
  	char	*str = argv[i++];
  	if (*str != '\0') {
		ret = tty_write(interp, argv[0], fileno(portfp), str);
		if (ret != TCL_OK) {
			return TCL_ERROR;
		}
  	}
  	if (i >= argc) {
  		break;
  	}
  	str = argv[i++];
  	if (*str == '\0') {
  		continue;
  	}
  	waitstr[0] = str;
  	found = tty_waitfor(portfd, count, waitidx, waitstr, timeout);
	if (found == 0) {
		continue;
	}
	if (found < 0) {
		sprintf(failmsg, "%s: timed out in chat script "
				 "while waiting for %s",
				 argv[0], str);
	}
	if (found > 0) {
		sprintf(failmsg, "%s: error in chat script, "
				 "got %s", argv[0], waitstr[found]);
	}
	if (sys_sigchat()) {
		Tcl_SetVar (interp, "chatFailCode", failmsg, TCL_LEAVE_ERR_MSG);
		return TCL_OK;
	}
	strcpy(interp->result, failmsg);
	return TCL_ERROR;
  }
  return TCL_OK;
}

/*==============================================================*
 * tty_write	write a string to the given tty, taking care
 *		of timeouts.
 *==============================================================*/
static int
tty_write(Tcl_Interp *interp, char *cmd, int portfd, char *sendstr)
{
  int	n;

  n = strlen(sendstr);
  signal(SIGALRM, get_alrm);
  tty_setalarm(2);
  if (write(portfd, sendstr, n) != n) {
  	if (errno == EINTR) {
  		sprintf(interp->result, "%s: timed out in write() call",
  					 cmd);
  	} else {
  		sprintf(interp->result, "%s: error in write() call: %s",
  					 cmd, strerror(errno));
  	}
  	tty_clralarm();
  	return TCL_ERROR;
  }
  tty_clralarm();
  return TCL_OK;
}


/*==============================================================*
 * tty_waitfor	wait for a set of specific responses.
 *==============================================================*/
static int
tty_waitfor(int portfd, int count, int *index, char **string, int timeout)
{
  int	j;
  char	c;

  for (j = 0; j < count; j++) {
  	index[j] = 0;
  }
  tty_setalarm(timeout);
  while (read(portfd, &c, 1) == 1) {
	for (j=0; j < count; j++) {
		char	*sp = string[j], c0;
		int	i, k = index[j];

		if (sp[k] == c) {
			if (sp[k+1] == '\0') {
  				tty_clralarm();
				return j;
			}
			index[j]++;
			continue;
		}
		c0 = sp[0];
		index[j] = 0;
		for (i=0; i <= k; i++) {
			if (sp[k-i] == c && !strncmp(sp, sp+i, k-i)) {
				index[j] = k-i+1;
				break;
			}
		}
	}
  }
  tty_clralarm();
  return -1;
}

/*==============================================================*
 * tty_setalarm	- set the alarm timer.
 * tty_clralarm	- reset the alarm timer.
 * get_alrm	- catch an alarm signal.
 *==============================================================*/
static void
tty_setalarm(int timeout)
{
  struct sigaction	act;

  act.sa_handler = get_alrm;
  sigemptyset(&act.sa_mask);
  act.sa_flags = 0;
  sigaction(SIGALRM, &act, &tty_sigoact);
}

static void
tty_clralarm(void)
{
  alarm(0);
  sigaction(SIGALRM, &tty_sigoact, NULL);
}

void
get_alrm(int signal)
{}
