/*==================================================================*
 * kern.c	- kernel interfacing; highly system-dependent!
 *
 * Coypright (C) 1994, 1995 by O.Kirch, okir@monad.swb.de
 * For copying conditions, see the file COPYING.
 *==================================================================*/
#include <linux/if.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <dirent.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "kern.h"

/* Indices of items in /proc/net/dev:
 *    ipacket ierrs idrop ififo iframe opacket oerrs odrop ofifo oframe 
 */
enum { DEV_IPACKET, DEV_IERRS, DEV_IDROP, DEV_IFIFO, DEV_IFRAME,
       DEV_OPACKET, DEV_OERRS, DEV_ODROP, DEV_OFIFO, DEV_OFRAME,
       DEV_NFIELDS
};
#define DEV_LAST	(DEV_NFIELDS-1)

/* List of flag names versus flag values
 */
static struct {
  short		value;
  char		*name;
} krn_fnames[] = {
  IFF_UP,		"UP",
  IFF_BROADCAST,	"BROADCAST",
  IFF_DEBUG,		"DEBUG",
  IFF_LOOPBACK,		"LOOPBACK",
  IFF_POINTOPOINT,	"POINTOPOINT",
  IFF_NOTRAILERS,	"NOTRAILERS",
  IFF_RUNNING,		"RUNNING",
  IFF_NOARP,		"NOARP",
  IFF_PROMISC,		"PROMISC",
  IFF_ALLMULTI,		"ALLMULTI",
  IFF_SLAVE,		"SLAVE",
  IFF_MASTER,		"MASTER",
  0,			NULL
};


/*==============================================================*
 * krn_getsockfd
 * Return a socket fd for a dgram socket. This socket is used
 * for various SIOC* calls. Returns -1 on error.
 *==============================================================*/
static int
krn_getsockfd(void)
{
  static int	sockfd = -2;

  if (sockfd == -2 && (sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
	sockfd = -1;
  }
  return sockfd;
}

/*==============================================================*
 * krn_pidbyname
 *==============================================================*/
int
krn_pidbyname(char *process)
{
  struct dirent	*dirent;
  char		fname[PATH_MAX], *sp;
  char		*s = NULL, *t = NULL;
  char		buffer[256];
  DIR		*dirp;
  FILE		*fp;
  int		ok, pid;

  if ((dirp = opendir("/proc")) == NULL) {
  	return -1;
  }
  sprintf(fname, "/proc/");
  sp = fname + strlen(fname);
  buffer[255] = '\0';
  while ((dirent = readdir(dirp)) != NULL) {
	sprintf (sp, "%s/stat", dirent->d_name);
	if (!isdigit(*sp) || (fp = fopen(fname, "r")) == NULL)
		continue;
  	ok = fgets(buffer, 255, fp) != NULL
  	 	&& (s = strchr(buffer, '(')) != NULL
  	 	&& (t = strchr(++s, ')')) != NULL;
  	fclose (fp);
  	if (!ok) continue;
  	*t = '\0';
  	if (!strcmp(process, s)) {
  		pid = atoi(dirent->d_name);
  		closedir(dirp);
  		return pid;
  	}
  }
  closedir(dirp);
  return -1;
}

/*==============================================================*
 * krn_iflist
 *==============================================================*/
char **
krn_iflist(int all_ifs)
{
  struct ifconf	ifc;
  struct ifreq	*ifr;
  char		buffer[1024];
  char **	iflist;
  int		ifct, ifmax;
  int		sockfd;
  int		i, len;

  ifc.ifc_len = sizeof(buffer);
  ifc.ifc_buf = buffer;

  sockfd = krn_getsockfd();
  if (sockfd < 0 || ioctl(sockfd, SIOCGIFCONF, &ifc) < 0) {
	return NULL;
  }

  ifr = ifc.ifc_req;
  len = ifc.ifc_len / sizeof(*ifr);
  iflist = NULL; ifct = ifmax = 0;
  for (i = 0; i < len; i++, ifr++) {
	if (!all_ifs) {
		ioctl(sockfd, SIOCGIFFLAGS, ifr);
		if (ioctl(sockfd, SIOCGIFFLAGS, ifr) < 0
		 || !(ifr->ifr_flags & IFF_UP)) {
			continue;
		}
	}
	if (ifmax <= ifct) {
		ifmax = ifct+4; /* heuristc: lo + eth0 + sl0 + NULL */
		iflist = (char **)realloc(iflist, ifmax * sizeof(char*));
		if (iflist == NULL) {
			return NULL;
		}
	}
	iflist[ifct++] = strdup (ifr->ifr_name);
  }
  if (ifmax <= ifct) {
	ifmax = ifct+1;
	iflist = (char **)realloc(iflist, ifmax * sizeof(char*));
	if (iflist == NULL) {
		return NULL;
	}
  }
  iflist[ifct++] = NULL;
  return iflist;
}

/*==============================================================*
 * krn_ifstats
 * Gather miscellaneous device info. The collection of parameters
 * is somewhat arbitrary, based on what I happened to need.
 *==============================================================*/
struct ifstat *
krn_ifstats(char *ifname)
{
  static struct ifstat ifs;
  struct ifreq	ifr;
  char		buffer[512];
  long		val[DEV_NFIELDS];
  FILE		*fp;
  char		*sp;
  int		sockfd;
  int		field;

  if ((fp = fopen("/proc/net/dev", "r")) == NULL) {
  	return NULL;
  }
  buffer[511] = '\0';
  fgets(buffer, 511, fp);	/* skip lines 1 & 2 */
  fgets(buffer, 511, fp);
  while (fgets(buffer, 511, fp) != NULL) {
	if ((sp = strtok(buffer, " \t:")) == NULL || strcmp(sp, ifname)) {
		continue;
	}
  	sp = strtok(NULL, " \t");
  	if (isdigit(*sp)) {
		for (field = 0; field < DEV_NFIELDS && sp != NULL; field++) {
			val[field] = atoi(sp);
			sp = strtok(NULL, " \t");
		}
		if (field < DEV_NFIELDS) {
			/* incompatible file format */
			fclose(fp);
			return NULL;
		}
	} else {
		/* no stats for this device... */
		for (field = 0; field < DEV_NFIELDS; field++) {
			val[field] = 0;
		}
	}
  	ifs.in.packets  = val[DEV_IPACKET];
  	ifs.out.packets = val[DEV_OPACKET];
  	ifs.in.errors   = val[DEV_IERRS];
  	ifs.out.errors  = val[DEV_OERRS];
  }
  fclose(fp);

  if ((sockfd = krn_getsockfd()) < 0) {
	return NULL;
  }

  strcpy (ifr.ifr_name, ifname);
  if (ioctl(sockfd, SIOCGIFFLAGS, &ifr) < 0) {
	return NULL;
  }
  ifs.flags = ifr.ifr_flags;

  strcpy (ifr.ifr_name, ifname);
  if (ioctl(sockfd, SIOCGIFADDR, &ifr) < 0) {
	return NULL;
  }
  ifs.addr = ifr.ifr_addr;

  strcpy (ifr.ifr_name, ifname);
  if (ioctl(sockfd, SIOCGIFDSTADDR, &ifr) < 0) {
	return NULL;
  }
  ifs.peeraddr = ifr.ifr_dstaddr;

  strcpy (ifr.ifr_name, ifname);
  if (ioctl(sockfd, SIOCGIFMTU, &ifr) < 0) {
	return NULL;
  }
  ifs.mtu = ifr.ifr_mtu;

  return &ifs;
}

/*==============================================================*
 * krn_ifflag_to_name
 *==============================================================*/
int
krn_ifflag_to_name(short flags, char *flagnames[KRN_MAX_IFFLAGS])
{
  short	flag;
  int	i, j = 0;

  for (i = 0; krn_fnames[i].value != 0; i++) {
	if ((flag = krn_fnames[i].value) & flags) {
		flagnames[j++] = krn_fnames[i].name;
	}
  }
  return j;
}

