/*================================================================*
 * modem.c	- POSIX termio.
 *
 * All functions were stolen from M.v. Smoorenburg's minicom.
 * I removed support for everything but plain POSIX. To support
 * a different system, simply write a different modem.c.
 * I also added functions to save a tty's state.
 * 						--okir
 *
 * Coypright (C) by Miquel van Smoorenburg
 * The functions in this file are covered by the GNU GPL.
 *================================================================*/
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <unistd.h>
#include <termios.h>
#include <stdio.h>
#include <setjmp.h>

#include "config.h"
#include "modem.h"

/*
 * Drop DTR line and raise it again.
 */
void
mdm_dtrtoggle(int fd) 
{
  struct termios tty, old;

  tcgetattr(fd, &tty);
  tcgetattr(fd, &old);
  cfsetospeed(&tty, B0);
  cfsetispeed(&tty, B0);
  tcsetattr(fd, TCSANOW, &tty);
  sleep(1);
  tcsetattr(fd, TCSANOW, &old);
}

/*
 * Send a break
 */
void
mdm_break(int fd)
{ 
  tcsendbreak(fd, 0);
}

/*
 * Get the dcd status
 */
int
mdm_getdcd(int fd)
{
  int mcs;
   
  ioctl(fd, TIOCMGET, &mcs);
  return(mcs & TIOCM_CAR ? 1 : 0);
}

/*
 * Save the state of a port
 */
void*
mdm_savestate(int fd)
{
  struct termios *savetty;

  if ((savetty = (struct termios*)malloc(sizeof(*savetty))) != NULL) {
  	tcgetattr(fd, savetty);
  }
  return (void*)savetty;
}

/*
 * Restore the state of a port
 */
void
mdm_restorestate(int fd, void* savetty)
{
  tcsetattr(fd, TCSANOW, (struct termios*)savetty);
}

/*
 * Set the line status so that it will not kill our process
 * if the line hangs up.
 */
void
mdm_nohang(int fd)
{
  struct termios sgg;
  
  tcgetattr(fd, &sgg);
  sgg.c_cflag |= CLOCAL;
  tcsetattr(fd, TCSANOW, &sgg);
}

/*
 * Set hangup on close on/off.
 */
void
mdm_hupcl(int fd, int on)
{
  struct termios sgg;
  
  tcgetattr(fd, &sgg);
  if (on)
  	sgg.c_cflag |= HUPCL;
  else
	sgg.c_cflag &= ~HUPCL;
  tcsetattr(fd, TCSANOW, &sgg);
}

/*
 * Flush the buffers
 */
void
mdm_flush(int fd, int input, int output)
{
  if (input && output) {
  	tcflush(fd, TCIOFLUSH);
  } else if (input) {
  	tcflush(fd, TCIFLUSH);
  } else if (output) {
  	tcflush(fd, TCOFLUSH);
  }
}

/*
 * Set baudrate, parity and number of bits.
 */
void
mdm_setparms(int fd, int speed, char parity, int bits)
{
  int spd = -1;
  struct termios tty;

  tcgetattr(fd, &tty);

  switch(speed) {
  case 0:	spd = B0;	break;
  case 300:	spd = B300;	break;
  case 600:	spd = B600;	break;
  case 1200:	spd = B1200;	break;
  case 2400:	spd = B2400;	break;
  case 4800:	spd = B4800;	break;
  case 9600:	spd = B9600;	break;
  case 19200:	spd = B19200;	break;
  case 38400:	spd = B38400;	break;
  /* we don't handle any speeds higher than 38K4 yet. Too lazy:-) */
  default:	spd = B9600;
  		fprintf( stderr, "internal: weird speed value %d in m_setparms\n",
  				  speed );
  }
  
  cfsetospeed(&tty, (speed_t)spd);
  cfsetispeed(&tty, (speed_t)spd);

  switch (bits) {
  case 5:
  	tty.c_cflag = (tty.c_cflag & ~CSIZE) | CS5;
  	break;
  case 6:
  	tty.c_cflag = (tty.c_cflag & ~CSIZE) | CS6;
  	break;
  case 7:
  	tty.c_cflag = (tty.c_cflag & ~CSIZE) | CS7;
  	break;
  case 8:
  default:
	tty.c_cflag = (tty.c_cflag & ~CSIZE) | CS8;
  	break;
  }		
  /* Set into raw, no echo mode */
  tty.c_iflag &= ~(IGNBRK | IGNCR | INLCR | ICRNL | IUCLC | 
  	IXANY | IXON | IXOFF | INPCK | ISTRIP);
  tty.c_iflag |= (BRKINT | IGNPAR);
  tty.c_oflag &= ~OPOST;
  tty.c_lflag = ~(ICANON | ISIG | ECHO | ECHONL | ECHOE | ECHOK);
  tty.c_cflag |= CREAD | CRTSCTS;
  tty.c_cc[VMIN] = 1;
  tty.c_cc[VTIME] = 5;

  tty.c_cflag &= ~(PARENB | PARODD);
  if (parity == 'E')
	tty.c_cflag |= PARENB;
  else if (parity == 'O')
	tty.c_cflag |= PARODD;

  tcsetattr(fd, TCSANOW, &tty);
}

/*
 * After a fork(), create a new session, and set its controlling tty
 * to the given one.
 */
void
mdm_daemon(int fd)
{
  setsid();
  if (fd >= 0) {
  	tcsetpgrp(fd, getpid());
  }
}

