/*==================================================================*
 * net.c	- network-specific stuff
 *
 * Coypright (C) 1994, 1995 by O.Kirch, okir@monad.swb.de
 * For copying conditions, see the file COPYING.
 *==================================================================*/
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <linux/if_slip.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <termios.h>
#include <signal.h>
#include <fcntl.h>
#include <limits.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>

#include <tcl/tcl.h>

#include "slipsh.h"
#include "kern.h"

static char *	net_printaddr(struct sockaddr * sa, int numeric);

/*==============================================================*
 * net_slip	convert line to slip mode
 * Usage:
 *	tty slip [-vj] [-compressed] [-6] [-6bit] tty
 *==============================================================*/
int
net_slip(ClientData cd, Tcl_Interp *interp, int argc, char *argv[])
{
  int		compress = 0;
  int		sixbit = 0;
  int		adaptive = 0;
  FILE		*portfp;
  int		portfd;
  int		ldisc, encap;
  char		ifname[32];
  int		i, ret;

  struct option options[] = {
	{ "6bit",	T_TRUE,		&sixbit			},
	{ "6",		T_TRUE,		&sixbit			},
	{ "adaptive",	T_TRUE,		&adaptive		},
	{ "compressed",	T_TRUE,		&compress		},
	{ "vj",		T_TRUE,		&compress		},
	{ NULL,		T_VOID,		NULL			}
  };

  /* Parse arguments */
  i = 1;
  if ((ret = tcl_getopt(interp, options, argc, argv, &i)) != TCL_OK) {
  	return TCL_ERROR;
  }
  if (i < argc-1) {
  	sprintf(interp->result, "%s: unknown argument: %s", argv[0], argv[i]);
  	return TCL_ERROR;
  }

  if (i != argc-1) {
  	sprintf(interp->result, "%s: need tty argument", argv[0]);
  	return TCL_ERROR;
  }
  if (Tcl_GetOpenFile(interp, argv[i++], 1, 1, &portfp) != TCL_OK) {
  	sprintf(interp->result, "%s: bad file argument", argv[0]);
  	return TCL_ERROR;
  }
  portfd = fileno(portfp);

  /* set line disc & encap */
  ldisc = N_SLIP;
  if (compress) {
  	encap = SL_MODE_CSLIP;
  } else {
  	if (adaptive) {
  		encap = SL_MODE_SLIP | SL_OPT_ADAPTIVE;
  	} else if (sixbit) {
  		encap = SL_MODE_SLIP | SL_OPT_SIXBIT;
  	} else {
  		encap = SL_MODE_SLIP;
  	}
  }
  if (ioctl(portfd, TIOCSETD, &ldisc) < 0) {
  	sprintf(interp->result, "%s: couldn't set line discipline N_SLIP",
  				 argv[0] );
  	return TCL_ERROR;
  }
  if (ioctl(portfd, SIOCSIFENCAP, &encap) < 0) {
  	sprintf(interp->result, "%s: couldn't set excapsulation proto %s",
  				 argv[0], compress? "CSLIP" : "SLIP" );
  	return TCL_ERROR;
  }
  if (ioctl(portfd, SIOCGIFNAME, ifname) < 0) {
  	sprintf(interp->result, "%s: couldn't get interface name",
  				 argv[0] );
  	return TCL_ERROR;
  }
  sprintf(interp->result, "%s", ifname);
  return TCL_OK;
}

/*==============================================================*
 * net_ifstats	get interface statistics
 * Usage:
 *	net ifstats 
 *		[-ipackets <varname>] [-ierrors <varname>]
 *		[-opackets <varname>] [-oerrors <varname>]
 *		[-address  <varname>] [-peeraddr <varname>]
 *		[-flags    <varname>]
 *		<iface>
 *
 * Get interface statistics.  In each of the variables specified,
 * return the corrsponding value for the interface.
 *
 * More items will be added as necessary.
 *==============================================================*/
int
net_ifstats(ClientData cd, Tcl_Interp *interp, int argc, char *argv[])
{
  char		*var_ipkt = NULL, *var_opkt = NULL;
  char		*var_ierr = NULL, *var_oerr = NULL;
  char		*var_addr = NULL, *var_peer = NULL;
  char		*var_flag = NULL;
  int		numeric = 1;
  struct ifstat	*ifstats;
  char		*ifname;
  char		val[12];
  int		i, ret;

  struct option options[] = {
	{ "ipackets",	T_STRING,	(void*)&var_ipkt	},
	{ "opackets",	T_STRING,	(void*)&var_opkt	},
	{ "ierrors",	T_STRING,	(void*)&var_ierr	},
	{ "oerrors",	T_STRING,	(void*)&var_oerr	},
	{ "address",	T_STRING,	(void*)&var_addr	},
	{ "peeraddress",T_STRING,	(void*)&var_peer	},
	{ "flags",	T_STRING,	(void*)&var_flag	},
	{ "resolve",	T_FALSE,	(void*)&numeric		},
	{ NULL,		T_VOID,		NULL			}
  };

  /* Parse arguments */
  i = 1;
  if ((ret = tcl_getopt(interp, options, argc, argv, &i)) != TCL_OK) {
  	return TCL_ERROR;
  }
  if (i < argc-1) {
  	sprintf(interp->result, "%s: unknown argument: %s", argv[0], argv[i]);
  	return TCL_ERROR;
  }

  if (i != argc-1) {
  	sprintf(interp->result, "%s: need interface argument", argv[0]);
  	return TCL_ERROR;
  }
  ifname = argv[i++];
  if ((ifstats = krn_ifstats(ifname)) == NULL) {
  	sprintf(interp->result, "%s: can't get interface statistics for %s",
  				 argv[0], ifname);
  	return TCL_ERROR;
  }
  if (var_ipkt != NULL) {
  	sprintf(val, "%d", ifstats->in.packets);
  	Tcl_SetVar(interp, var_ipkt, val, 0);
  }
  if (var_opkt != NULL) {
  	sprintf(val, "%d", ifstats->out.packets);
  	Tcl_SetVar(interp, var_opkt, val, 0);
  }
  if (var_ierr != NULL) {
  	sprintf(val, "%d", ifstats->in.errors);
  	Tcl_SetVar(interp, var_ierr, val, 0);
  }
  if (var_oerr != NULL) {
  	sprintf(val, "%d", ifstats->out.errors);
  	Tcl_SetVar(interp, var_oerr, val, 0);
  }

  /* Translate the array of flag names into a list.
   */
  if (var_flag != NULL) {
	char		*fnames[KRN_MAX_IFFLAGS];
	int		flagct;

	flagct = krn_ifflag_to_name(ifstats->flags, fnames);
	Tcl_SetVar(interp, var_flag, "", 0);
	for (i = 0; i < flagct; i++) {
		Tcl_SetVar(interp, var_flag, fnames[i],
				TCL_APPEND_VALUE|TCL_LIST_ELEMENT);
	}
  }
  if (var_addr != NULL) {
	Tcl_SetVar(interp, var_addr,
			net_printaddr(&ifstats->addr, numeric), 0);
  }
  if (var_peer != NULL) {
	Tcl_SetVar(interp, var_peer,
			net_printaddr(&ifstats->peeraddr, numeric), 0);
  }
  return TCL_OK;
}

/*==============================================================*
 * net_iflist	get list of network interfaces
 * Usage:
 *	net iflist [-all]
 *
 * Get interface list. Returns a TCL list containing the list
 * of configured interfaces.
 *==============================================================*/
int
net_iflist(ClientData cd, Tcl_Interp *interp, int argc, char *argv[])
{
  char		**iflist;
  int		opt_all = 0;
  int		i, ret;

  struct option options[] = {
	{ "all",	T_TRUE,		&opt_all		},
	{ NULL,		T_VOID,		NULL			}
  };

  /* Parse arguments */
  i = 1;
  if ((ret = tcl_getopt(interp, options, argc, argv, &i)) != TCL_OK) {
  	return TCL_ERROR;
  }
  if (i < argc) {
  	sprintf(interp->result, "%s: unknown argument: %s", argv[0], argv[i]);
  	return TCL_ERROR;
  }

  if ((iflist = krn_iflist(opt_all)) == NULL) {
  	sprintf(interp->result, "%s: couldn't read interface list", argv[0]);
	return TCL_ERROR;
  }

  Tcl_ResetResult (interp);
  for (i = 0; iflist[i] != NULL; i++) {
	Tcl_AppendElement (interp, iflist[i]);
	free (iflist[i]);
  }
  free (iflist);

  return TCL_OK;
}

/*==============================================================*
 * print network address. We handle only AF_INET for the moment.
 * We also don't build a local cache of resolved addrs here, 
 * although that might be useful. I hope the resolver handles
 * that...
 *==============================================================*/
static char *
net_printaddr(struct sockaddr *sa, int numeric)
{
  static char		buffer[258];
  struct sockaddr_in *	sin;
  unsigned long		addr;

  buffer[0] = '\0';

  sin = (struct sockaddr_in *)sa;
  if (sin->sin_family != AF_INET) {
	strcpy (buffer, "[NONE SET]");
	return buffer;
  }

  addr = (unsigned long)(sin->sin_addr.s_addr);
  if (addr == INADDR_ANY) {
	if (!numeric) {
		strcpy (buffer, "default");
	} else {
		strcpy (buffer, inet_ntoa(sin->sin_addr));
	}
	return buffer;
  }
  if (!numeric) {
	struct hostent	*hp;

	hp = gethostbyaddr((char*)&addr, sizeof(addr), AF_INET);
	if (hp != NULL) {
		strcpy (buffer, hp->h_name);
	}
  }
  if (buffer[0] == '\0') {
	strcpy (buffer, inet_ntoa(sin->sin_addr));
  }
  return buffer;
}
