################################################################## 
# common slipsh routines for managing a TCL link
##################################################################
set pidfile		"/tmp/slip"
set logfile		"/var/log/slipstat"

##################################################################
# Dial up a remote system and log in
##################################################################
proc dialup {device ttyfd telnolist login passwd retries} {

    global  chatFailCode
    global  chat_ok

    sys signal CHATFAIL "set chat_ok 0;"

    for {set attempt 1} {$attempt <= $retries} {incr attempt} {
	foreach telno $telnolist {
            set chat_ok 1
	    puts -nonewline stderr "dialing $telno... "; flush stderr
            tty chat -error "ERROR\r" -error "BUSY\r" \
                     -error "NO DIALTONE\r" -error "NO CARRIER\r" \
                     -timeout 120 $ttyfd \
                     ATZ\r OK\r ATL1M1\r OK\r ATDT$telno\r CONNECT

	    if {$chat_ok == 0} {
	        puts stderr "failed: $chatFailCode"
		sys sleep 2
	    } else {
                tty recv $ttyfd speed
                puts stderr "connected with $speed"

	        puts -nonewline stderr "logging in... "; flush stderr

                tty chat -error "NO CARRIER\r" -timeout 120 $ttyfd \
		         "" ogin: $login\r ssword: $passwd\r

                if {$chat_ok != 0} {
                    puts stderr "OK."
                    return "";
	        }
	        puts stderr "failed: $chatFailCode"
	        puts stderr "hanging up"
	        tty hangup $ttyfd
	    }
	}
	if {$attempt < $retries} {
	    puts stderr "retrying in 60 seconds... "
	    sys sleep 60
	    puts stderr "retry $attempt of $retries";
	}
    }

    puts stderr "giving up."
    return "FAILED";
}

##################################################################
# Convert a tty to CSLIP, and set up the standard routes.
##################################################################
proc setslip {mode ttyfd local remote} {

    switch $mode {
    SLIP     { set iface [net slip $ttyfd]; }
    CSLIP    { set iface [net slip -compressed $ttyfd]; }
    ADAPTIVE { set iface [net slip -adaptive $ttyfd]; }
    SLIP6    { set iface [net slip -6bit $ttyfd]; }
    *        { puts stderr "Warning: no mode selected in setslip";
               set iface [net slip $ttyfd];
             }
    }

    puts "started SLIP link $local --> $remote; interface $iface"
    ifconfig $iface $local pointopoint $remote

    # this notifies the slip monitor script (if running) that the
    # interface has been activated. Ignore any error returns.
    catch exec {sys kill SIGHUP slipmon}

    return $iface
}

##################################################################
# Invoke the ifconfig command. The extra eval is necessary to break
# up the Tcl list from $args into separate arguments.
##################################################################
proc ifconfig args {

    eval exec /sbin/ifconfig $args
    return "";
}

##################################################################
# Invoke the route command. The extra eval is necessary to break
# up the Tcl list from $args into separate arguments
##################################################################
proc route args {

    eval exec /sbin/route $args
    return "";
}

##################################################################
# Become a daemon - fork, and make the given tty the child's controlling
# tty. Recreate the tty lockfile so it contains the child's PID (This
# is an important detail of the file locking protocol one tends to forget
# easily :-).
##################################################################
proc daemon {device ttyfd} {

    sys signal SIGHUP "hangup $ttyfd $device"
    close stdin;

    # fork, and make the SLIP link the child's controlling tty
    set pid [sys fork -tty $ttyfd]
    if {$pid != 0} {
        tty lock -pid $pid $device
    }
    return $pid
}

##################################################################
# Hangup the serial line, and notify the slip monitor of this fact
##################################################################
proc hangup {ttyfd device} {

    catch {sys kill SIGHUP slipmon}

    tty close  $ttyfd
    tty unlock $device

    catch {sys kill SIGHUP slipmon}

    exit 0
}

##################################################################
# Write our PID to a ``lock'' file. It's doesn't really lock anything, 
# this is only used when someone wants to hangup the SLIP link.
##################################################################
proc log_pid {pid remote iface} {

    global pidfile

    set fd [open "$pidfile.$remote" "w"];
    puts $fd $pid
    close $fd
}

proc remove_pid {remote} {

    global pidfile

    exec /bin/rm -f "$pidfile.$remote"
}

##################################################################
# Get the pid associated with a SLIP link
##################################################################
proc get_pid {remote} {

    global pidfile

    set file "$pidfile.$remote"

    if {![file exists $file]} {
	return "";
    }

    set fd [open $file];
    set pid [gets $fd]
    close $fd
    return "$pid";
}

##################################################################
# Log a message to the SLIP log file
##################################################################
proc sliplog {msg} {

    global logfile

    if {[file exists $logfile]} {
	exec /bin/date "+%b %d %T $msg" >> $logfile
    }
}
