/*==================================================================*
 * signal.c	send and catch signals etc.
 *
 * Coypright (C) 1994, 1995 by O.Kirch, okir@monad.swb.de
 * For copying conditions, see the file COPYING.
 *==================================================================*/
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <limits.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <tcl/tcl.h>

#include "slipsh.h"
#include "kern.h"

static void	sys_sigcatcher(int sig);
static void	sys_sigtracer(ClientData cd, Tcl_Interp *interp,
				int level, char *cmd, Tcl_CmdProc *exec,
				ClientData cmdCd, int argc, char **argv);

static struct sys_signame {
	char		*name;
	int		signo;
} signames[] = {
      { "CHATFAIL",	0	   },
      { "SIGALRM",	SIGALRM	   },
      { "SIGBUS",	SIGBUS	   },
      { "SIGCHLD",	SIGCHLD	   },
      { "SIGFPE",	SIGFPE	   },
      { "SIGHUP",	SIGHUP	   },
      { "SIGILL",	SIGILL	   },
      { "SIGINT",	SIGINT	   },
      { "SIGIOT",	SIGIOT	   },
      { "SIGKILL",	SIGKILL	   },
      { "SIGPIPE",	SIGPIPE	   },
      { "SIGPROF",	SIGPROF	   },
      { "SIGQUIT",	SIGQUIT	   },
      { "SIGSEGV",	SIGSEGV	   },
      { "SIGTERM",	SIGTERM	   },
      { "SIGTRAP",	SIGTRAP	   },
      { "SIGUSR1",	SIGUSR1	   },
      { "SIGUSR2",	SIGUSR2	   },
      { "SIGVTALRM",	SIGVTALRM  },
      { NULL,		0	   }
};

static struct sys_siginfo {
	Tcl_Interp	*interp;
	char		*command;
	short		pending;
	short		installed;
	struct sigaction oldact;
} signals[NSIG];
static int		sys_sigcount = 0;
static int		initialized = 0;
static Tcl_Trace	sigtrace_handle;

/*==============================================================*
 * sys_siginit	- initialize signal handling
 *==============================================================*/
static void
sys_siginit(Tcl_Interp *interp)
{
  int		i;

  for (i=0; i < NSIG; i++) {
  	signals[i].interp = NULL;
  	signals[i].command = NULL;
  	signals[i].pending = 0;
  	signals[i].installed = 0;
  }
  sigtrace_handle = Tcl_CreateTrace(interp, 1024, sys_sigtracer, NULL);

  initialized = 1;
}

/*==============================================================*
 * sys_sigtracer: we execute signals asynchronuously by tracing
 * TCL execution, and fire the signal handlers whenever a signal
 * occurred.
 *==============================================================*/
void
sys_sigtracer(ClientData cd, Tcl_Interp *interp, int level,
	char *cmd, Tcl_CmdProc *exec, ClientData cmdCd, int argc, char **argv)
{
  struct sys_siginfo *sigp;
  static int	  recur = 0;
  int		  signo = 0;
  int		  ret;

  if (!sys_sigcount || recur) {
  	return;
  }
  recur = 1;

  /* we do a round-robin scheduling of signals here for no special
   * reason.
   */
  while (sys_sigcount) {
  	sigp = &signals[signo];
  	if (sigp->pending) {
		if (sigp->command != NULL) {
			ret = Tcl_Eval(sigp->interp, sigp->command);
			if (ret == TCL_ERROR) {
				fprintf (stderr, "slipsh: signal handler "
						 "returned error: %s\n",
							interp->result);
			}
		} else {
			fprintf (stderr, "slipsh: internal error! "
					 "no command for caught signal %d?!\n",
					 signo);
		}
		sys_sigcount--;
		sigp->pending--;
  	}
  	if (++signo >= NSIG) {
  		signo = 0;
  	}
  }
  recur = 0;
}

/*==============================================================*
 * sys_signal	- install or obtain a signal handler
 * Usage:
 *	sys signal signame command
 *
 * signame must be one of the signal names defined in the
 * array at the top of this file. 
 * SIG_DFL and SIG_IGN aren't handled yet.
 *==============================================================*/
int
sys_signal(ClientData cd, Tcl_Interp *interp, int argc, char *argv[])
{
  struct sys_siginfo *sigp;
  char		  *signame;
  int		  signo;
  int		  i, j, ret;

  struct option options[] = {
	{ NULL,		T_VOID,		NULL			}
  };

  if (!initialized) {
  	sys_siginit(interp);
  }

  /* Parse arguments */
  i = 1;
  if ((ret = tcl_getopt(interp, options, argc, argv, &i)) != TCL_OK) {
  	return TCL_ERROR;
  }
  if (i < argc-2) {
  	sprintf(interp->result, "%s: unknown argument: %s", argv[0], argv[i]);
  	return TCL_ERROR;
  }
  if (i != argc-2) {
  	sprintf(interp->result, "%s: usage: %s signame command",
  				argv[0], argv[0]);
  	return TCL_ERROR;
  }

  signame = argv[i++];
  for (j = 0, signo = -1; signames[j].name != NULL; j++) {
  	if (!strcasecmp(signame, signames[j].name)) {
  		signo = signames[j].signo;
  		break;
  	}
  }
  if (signo == -1) {
  	sprintf(interp->result, "%s: no such signal: %s", argv[0], signame);
  	return TCL_ERROR;
  }
  sigp = &signals[signo];
  if (sigp->pending) {
  	sprintf(interp->result, "%s: signal %s pending - can't install "
  				"new handler", argv[0], signame);
  	return TCL_ERROR;
  }
  if (sigp->command != NULL) {
	/* for once, we use the clumsy Tcl_SetResult instead of the
	 * usual sprintf, because the signal handler may actually be
	 * very big... */
	Tcl_SetResult(interp, sigp->command, TCL_VOLATILE);
  	free (sigp->command);
  }
  sigp->interp  = interp;
  sigp->command = strdup(argv[i]);
  sigp->pending = 0;

  if (!sigp->installed) {
  	struct sigaction act;

  	sigemptyset(&act.sa_mask);
  	act.sa_handler = sys_sigcatcher;
  	act.sa_flags = 0;
  	act.sa_restorer = NULL;
  	sigaction(signo, &act, &sigp->oldact);
  	sigp->installed = 1;
  }

  return TCL_OK;
}

/*==============================================================*
 * sys_kill	- send a signal to a process
 * Usage:
 *	sys kill signame process
 *==============================================================*/
int
sys_kill(ClientData cd, Tcl_Interp *interp, int argc, char *argv[])
{
  char		*signame;
  char		*process;
  int		signo;
  int		i, j, n, ret;
  int		pid;

  struct option options[] = {
	{ NULL,		T_VOID,		NULL			}
  };

  if (!initialized) {
  	sys_siginit(interp);
  }

  /* Parse arguments */
  i = 1;
  if ((ret = tcl_getopt(interp, options, argc, argv, &i)) != TCL_OK) {
  	return TCL_ERROR;
  }
  if (i < argc-2) {
  	sprintf(interp->result, "%s: unknown argument: %s", argv[0], argv[i]);
  	return TCL_ERROR;
  }
  if (i != argc-2) {
  	sprintf(interp->result, "%s: usage: %s signame process",
  				argv[0], argv[0]);
  	return TCL_ERROR;
  }

  signame = argv[i++];
  for (j = 0, signo = -1; signames[j].name != NULL; j++) {
  	if (!strcasecmp(signame, signames[j].name)) {
  		signo = signames[j].signo;
  		break;
  	}
  }
  if (signo == -1) {
  	sprintf(interp->result, "%s: no such signal: %s", argv[0], signame);
  	return TCL_ERROR;
  }
  process = argv[i++];
  n = strspn(process, "0123456789");
  if (process[n] != '\0') {
  	if ((pid = krn_pidbyname(process)) < 0) {
  		sprintf(interp->result, "%s: no such process: \"%s\"", 
  					argv[0], process);
		return TCL_ERROR;
  	}
  } else {
  	pid = atoi(process);
  }
  if (pid == 1) {
  	sprintf(interp->result, "%s: sorry, won't let you kill pid %d",
  				argv[0], pid);
  	return TCL_ERROR;
  }
  if (kill(pid, signo) < 0) {
  	sprintf(interp->result, "%s: can't kill %d: %s",
  				argv[0], pid, strerror(errno));
  	return TCL_ERROR;
  }
  return TCL_OK;
}

/*==============================================================*
 * sys_sigchat	- handle chat script failures.  If a handler
 *		for chat failures is installed, a signal condi-
 *		tion is marked, and the function returns 1.
 *		Otherwise, we return 0.
 *==============================================================*/
int
sys_sigchat(void)
{
  if (!signals[0].installed) {
  	return 0;
  }
  signals[0].pending++;
  sys_sigcount++;
  return 1;
}

/*==============================================================*
 * sys_sigcatcher	- catch signals and record them
 *==============================================================*/
static void
sys_sigcatcher(int sig)
{
  signals[sig].pending++;
  sys_sigcount++;
}
