/*==================================================================*
 * sys.c	- miscellaneous system calls
 *
 * Coypright (C) 1994, 1995 by O.Kirch, okir@monad.swb.de
 * For copying conditions, see the file COPYING.
 *==================================================================*/
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <limits.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>

#include <tcl/tcl.h>

#include "slipsh.h"
#include "modem.h"

/*==============================================================*
 * sys_fork	fork a child process
 * Usage:
 *	sys fork [-tty fd] [-daemon]
 * Fork off a child process, optionally setting its controlling
 * tty to fd. The daemon option tells it to detach itself from
 * the terminal without setting any controlling tty.
 *
 * If successful, the comand returns the pid of the child (parent)
 * or 0 (child).
 *==============================================================*/
int
sys_fork(ClientData cd, Tcl_Interp *interp, int argc, char *argv[])
{
  char		*tty = NULL;
  int		daemon = 0;
  FILE		*portfp;
  int		portfd = -1;
  int		i, ret;
  int		pid;

  struct option options[] = {
	{ "tty",	T_STRING,	&tty			},
	{ "daemon",	T_TRUE,		&daemon			},
	{ NULL,		T_VOID,		NULL			}
  };

  /* Parse arguments */
  i = 1;
  if ((ret = tcl_getopt(interp, options, argc, argv, &i)) != TCL_OK) {
  	return TCL_ERROR;
  }
  if (i < argc) {
  	sprintf(interp->result, "%s: unknown argument: %s", argv[0], argv[i]);
  	return TCL_ERROR;
  }

  if (tty != NULL) {
  	if (Tcl_GetOpenFile(interp, tty, 1, 1, &portfp) != TCL_OK) {
  		sprintf(interp->result, "%s: bad tty argument", argv[0]);
  		return TCL_ERROR;
  	}
  	portfd = fileno(portfp);
  }

  if ((pid = fork()) < 0) {
  	sprintf(interp->result, "%s: error: %s", argv[0], strerror(errno));
  	return TCL_ERROR;
  }
  if (portfd >= 0 || daemon) {
  	mdm_daemon(portfd);
  }
  sprintf(interp->result, "%d", pid);
  return TCL_OK;
}

/*==============================================================*
 * sys_pause	pause - wait until signalled
 * Usage:
 *	sys pause
 *==============================================================*/
int
sys_pause(ClientData cd, Tcl_Interp *interp, int argc, char *argv[])
{
  int		i, ret;

  struct option options[] = {
	{ NULL,		T_VOID,		NULL			}
  };

  /* Parse arguments */
  i = 1;
  if ((ret = tcl_getopt(interp, options, argc, argv, &i)) != TCL_OK) {
  	return TCL_ERROR;
  }
  if (i < argc) {
  	sprintf(interp->result, "%s: unknown argument: %s", argv[0], argv[i]);
  	return TCL_ERROR;
  }
  pause();
  return TCL_OK;
}

/*==============================================================*
 * sys_sleep	sleep for some time
 * Usage:
 *	sys sleep <seconds>
 *==============================================================*/
int
sys_sleep(ClientData cd, Tcl_Interp *interp, int argc, char *argv[])
{
  int		interval;
  int		i, ret;

  static struct option options[] = {
	{ NULL,		T_VOID,		NULL			}
  };

  /* Parse arguments */
  i = 1;
  if ((ret = tcl_getopt(interp, options, argc, argv, &i)) != TCL_OK) {
  	return TCL_ERROR;
  }
  if (i < argc-1) {
  	sprintf(interp->result, "%s: unknown argument: %s", argv[0], argv[i]);
  	return TCL_ERROR;
  }
  if (i != argc-1) {
  	sprintf(interp->result, "%s: need time argument", argv[0] );
  	return TCL_ERROR;
  }
  interval = atoi(argv[i++]);
  sleep(interval);
  return TCL_OK;
}

/*==============================================================*
 * sys_time	get the current time in sec since 1970 Jan 1
 * Usage:
 *	sys time
 *==============================================================*/
int
sys_time(ClientData cd, Tcl_Interp *interp, int argc, char *argv[])
{
  time_t	now;
  int		i, ret;

  static struct option options[] = {
	{ NULL,		T_VOID,		NULL			}
  };

  /* Parse arguments */
  i = 1;
  if ((ret = tcl_getopt(interp, options, argc, argv, &i)) != TCL_OK) {
  	return TCL_ERROR;
  }
  if (i < argc) {
  	sprintf(interp->result, "%s: unknown argument: %s", argv[0], argv[i]);
  	return TCL_ERROR;
  }
  time (&now);
  sprintf (interp->result, "%ld", now);
  return TCL_OK;
}

