/*==================================================================*
 * tty.c	- this is the tty handling part of slide.
 *
 * Most of the TTY handling code in this file was `inspired'
 * by Miquel van Smoorenburg's minicom.
 *
 * Coypright (C) 1994, 1995 by O.Kirch, okir@monad.swb.de
 * For copying conditions, see the file COPYING.
 *==================================================================*/
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <termios.h>
#include <signal.h>
#include <fcntl.h>
#include <limits.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>

#include <tcl/tcl.h>

#include "slipsh.h"
#include "modem.h"

#ifndef HAVE_BASENAME
static char *	basename(char *s);
#endif
static char *	devname(char *dev);
static char *	lockname(char *lockdir, char *port);
static int	tty_savestate(int portfd);
static int	tty_restorestate(int portfd);
static int	tty_updatestate(int portfd, int newfd);
static int	tty_uid(char *user);
static int	tty_gid(char *group);

/*
 * local variables and type definitions
 */
static struct tty_state {
	struct tty_state	*next;
	int			portfd;
	int			uid, gid;
	int			mode;
	int			ldisc, encap;
	void			*state;
} *tty_state = NULL;


/*==============================================================*
 * tty_lock	create a lock file for the given tty.
 * Usage:
 *	tty lock [-ascii] [-binary] [-pid <pid>]
 *		[-directoy <dir>] [-shy]
 *		<device>
 * The ascii/binary options select whether the pid is written as
 * as 4byte binary, or in textual representation. The directory
 * option lets you specify the directory where lock files reside.
 * By default, this function removes any stale lockfiles. You can
 * override this by specifying the -shy option.
 *==============================================================*/
int
tty_lock(ClientData cd, Tcl_Interp *interp, int argc, char *argv[])
{
  char		*lockdir = NULL;
  int		lockpid = -1;
  int		brute = 1;
  char		*lockfile;
  char		*port;
  char		buf[128];
  int		lockascii = 1;
  int		fd, i, n = 0, ret;
  int		pid, mypid;

  struct option	options[] = {
	{ "ascii",	T_TRUE,		(void*)&lockascii	},
	{ "binary",	T_FALSE,	(void*)&lockascii	},
	{ "directory",	T_STRING,	(void*)&lockdir		},
	{ "pid",	T_INT,		(void*)&lockpid		},
	{ "shy",	T_FALSE,	(void*)&brute		},
	{ NULL,		T_VOID,		NULL			}
  };


  /* Parse arguments */
  i = 1;
  if ((ret = tcl_getopt(interp, options, argc, argv, &i)) != TCL_OK) {
  	return TCL_ERROR;
  }
  if (i < argc-1) {
  	sprintf(interp->result, "%s: unknown option %s",
  					argv[0], argv[i]);
  	return TCL_ERROR;
  }

  /* Make sure we have a tty name */
  if (i != argc-1) {
  	sprintf(interp->result, "%s: need tty_name argument", argv[0]);
  	return TCL_ERROR;
  }

  /* Construct name of device file */
  port = devname(argv[i]);
  lockfile = lockname(lockdir, port);
  mypid = getpid();

  /*
   * Check if a lock exists, and get pid of process who wrote the lock.
   * The lock is overridden, if we own it ourselves, or if the process
   * owning is has died.
   */
  if ((fd = open(lockfile, O_RDONLY)) >= 0) {
	n = read(fd, buf, 127);
	close(fd);
	if (n > 0) {
		pid = -1;
		if (n == sizeof(int)) {
			/* Binary lockfile. */
			pid = *(int *)buf;
		} else {
			/* Ascii lockfile. */
			sscanf(buf, "%d", &pid);
		}
		if (pid > 0 && kill(pid, 0) < 0 && errno == ESRCH && brute) {
		    fprintf(stderr, "Stale lockfile found. Removing it.\n");
		    unlink(lockfile);
		} else if (pid == mypid) {
		    unlink(lockfile);
		} else {
		    n = 0;
		}
	}
	if (n == 0) {
  		sprintf(interp->result, "%s: device %s locked.",
  					 argv[0], port);
  		return TCL_ERROR;
	}
  }
  if ((fd = open(lockfile, O_WRONLY | O_CREAT | O_EXCL, 0666)) < 0) {
  	sprintf(interp->result, "%s: cannot create lockfile %s",
  				 argv[0], lockfile);
  	return TCL_ERROR;
  }
  if (lockpid <= 0) {
  	lockpid = getpid();
  }
  if (lockascii) {
  	sprintf(buf, "%06d", lockpid);
  	i = strlen(buf);
  } else {
  	i = sizeof(int);
  	memcpy(buf, &lockpid, i);
  }
  if (write(fd, buf, i) != i) {
  	sprintf(interp->result, "%s: cannot write pid to lockfile %s",
  				 argv[0], lockfile);
  	unlink(lockfile);
  	close(fd);
  	return TCL_ERROR;
  }
  close(fd);
  sprintf(interp->result, "%s", lockfile);
  return TCL_OK;
}

/*==============================================================*
 * tty_unlock	remove a device lockfile
 * Usage:
 *	tty unlock [-directoy <dir>] device
 * The -directory option specifies the location of device lock
 * files.
 *==============================================================*/
int
tty_unlock(ClientData cd, Tcl_Interp *interp, int argc, char *argv[])
{
  char		*lockdir = NULL;
  char		*lockfile;
  char		*port;
  int		ret, i;

  struct option	options[] = {
	{ "directory",	T_STRING,	(void*)&lockdir		},
	{ NULL,		T_VOID,		NULL			}
  };


  /* Parse arguments */
  i = 1;
  if ((ret = tcl_getopt(interp, options, argc, argv, &i)) != TCL_OK) {
  	return TCL_ERROR;
  }
  if (i < argc-1) {
  	sprintf(interp->result, "%s: unknown option %s",
  					argv[0], argv[i]);
  	return TCL_ERROR;
  }

  /* Make sure we have a tty name */
  if (i != argc-1) {
  	sprintf(interp->result, "%s: need tty_name argument", argv[0]);
  	return TCL_ERROR;
  }
  port = devname(argv[1]);
  lockfile = lockname(lockdir, port);
  unlink(lockfile);
  return TCL_OK;
}

static void
get_alrm(int signal)
{}

/*==============================================================*
 * tty_open	open the given tty.
 * Usage:
 *	tty open [-bits (8|7|6)] [-parity (none|even|odd)]
 *		[-speed	<speed>] [-baud	<speed>]
 *		[-user <user>] [-group <group>]
 *		[-uid <uid>] [-gid <gid>] [-mode <mode>] device
 *
 * The -baud option is an alias for -speed.
 * uid/gid specifies the UID and GID the device will be chmod()ed
 * to. Alternatively, the user and group name can be given using
 * the user/group options.
 *==============================================================*/
int
tty_open(ClientData cd, Tcl_Interp *interp, int argc, char *argv[])
{
  char		*parity = "none";
  char		*user = NULL, *group = NULL;
  int		speed = 38400;
  int		bits = 8;
  int		dev_uid = -1, dev_gid = -1, dev_mode = -1;
  char		*port;
  FILE		*portfp;
  int		i, ret, fd, portfd;
  char		par;

  struct option options[] = {
	{ "baud",	T_INT,		&speed		},
	{ "bits",	T_INT,		&bits		},
	{ "parity",	T_STRING,	&parity		},
	{ "speed",	T_INT,		&speed		},
	{ "uid",	T_INT,		&dev_uid	},
	{ "gid",	T_INT,		&dev_gid	},
	{ "user",	T_STRING,	&user		},
	{ "group",	T_STRING,	&group		},
	{ "mode",	T_INT,		&dev_mode	},
	{ NULL,		T_VOID,		NULL		}
  };


  /* Parse arguments */
  i = 1;
  if ((ret = tcl_getopt(interp, options, argc, argv, &i)) != TCL_OK) {
  	return TCL_ERROR;
  }
  if (i < argc-1) {
  	sprintf(interp->result, "%s: unknown option %s",
  					argv[0], argv[i]);
  	return TCL_ERROR;
  }

  /* Make sure we have a tty name */
  if (i != argc-1) {
  	sprintf(interp->result, "%s: need tty_name argument", argv[0]);
  	return TCL_ERROR;
  }

  /* Construct name of device file */
  port = devname(argv[i]);

  /* parse parity value */
  if (!strcmp(parity, "none")) {
  	par = 'N';
  } else if (!strcmp(parity, "even")) {
  	par = 'E';
  } else if (!strcmp(parity, "odd")) {
  	par = 'O';
  } else {
  	sprintf(interp->result, "%s: parity must be none, even or odd",
  				 argv[0]);
  	return TCL_ERROR;
  }

  if (user != NULL && (dev_uid = tty_uid(user)) < 0) {
	sprintf(interp->result, "%s: no such user: %s", argv[0], user);
	return TCL_ERROR;
  }
  if (group != NULL && (dev_gid = tty_gid(group)) < 0) {
	sprintf(interp->result, "%s: no such group: %s", argv[0], group);
	return TCL_ERROR;
  }

  /*
   * Open the tty. We first open it with O_NDELAY.
   */
  signal(SIGALRM, get_alrm);
  alarm(2);
  portfd = open(port, O_RDWR|O_NDELAY);
  if (portfd >= 0) {
	/* Open it a second time, now without O_NDELAY. */
	if (tty_savestate(portfd) != TCL_OK) {
		sprintf(interp->result, "%s: couldn't save tty state", argv[0]);
		close(portfd);
		return TCL_ERROR;
	}
	mdm_setparms(portfd, speed, par, bits);
	fd = portfd;
	portfd = open(port, O_RDWR);
	if (portfd < 0) {
		/* restore tty state, ignoring errors */
		tty_restorestate(fd);
	}
	tty_updatestate(fd, portfd);
	close(fd);
  }
  alarm(0);
  signal(SIGALRM, SIG_IGN);
  if (portfd < 0) {
	sprintf(interp->result, "%s: can't open port %s: %s",
				 argv[0], port, strerror(errno));
	return TCL_ERROR;
  }

  /* Set device file owner and mode. A uid/gid of -1 means don't
   * change the file's user/group owner
   */
  chown(port, dev_uid, dev_gid);
  if (dev_mode != -1) {
  	chmod(port, dev_mode);
  }

  /* Set CLOCAL mode, HUPCL (in case program crashes), and flush. */
  mdm_nohang(portfd);
  mdm_hupcl(portfd, 1);
  mdm_flush(portfd, 1, 1);

  /* nonsense: sprintf(interp->result, "%s", port); */
  if ((portfp = fdopen(portfd, "r+")) == NULL) {
  	sprintf(interp->result, "%s: couldn't fdopen file", argv[0]);
  	close(portfd);
  	return TCL_ERROR;
  }
  Tcl_EnterFile(interp, portfp, TCL_FILE_READABLE|TCL_FILE_WRITABLE);
  return TCL_OK;
}

/*==============================================================*
 * tty_close		close a tty device
 * Usage:
 *	tty close [-norestore] <device>
 * The -norestore option prevents the function from restoring
 * the tty state saved when opening the device.
 *==============================================================*/
int
tty_close(ClientData cd, Tcl_Interp *interp, int argc, char *argv[])
{
  int		restore = 1;
  int		i, ret;
  FILE		*fp;

  struct option options[] = {
	{ "norestore",	T_FALSE,	(void*)&restore		},
	{ NULL,		T_VOID,		NULL			}
  };


  /* Parse arguments */
  i = 1;
  if ((ret = tcl_getopt(interp, options, argc, argv, &i)) != TCL_OK) {
  	return TCL_ERROR;
  }
  if (i < argc-1) {
  	sprintf(interp->result, "%s: unknown option %s",
  					argv[0], argv[i]);
  	return TCL_ERROR;
  }

  /* Make sure we have a tty name */
  if (i != argc-1) {
  	sprintf(interp->result, "%s: need tty_name argument", argv[0]);
  	return TCL_ERROR;
  }

  if (Tcl_GetOpenFile(interp, argv[1], 1, 0, &fp) != TCL_OK) {
  	sprintf(interp->result, "%s: no such file", argv[0]);
  	return TCL_ERROR;
  }
  if (restore) {
  	tty_restorestate(fileno(fp));
  }
  if (fclose(fp) != 0) {
  	sprintf(interp->result, "%s: failed to close file: %s",
  				 argv[1], strerror(errno));
  	return TCL_ERROR;
  }
  return TCL_OK;
}

/*==============================================================*
 * tty_flush		flush a tty device
 * Usage:
 *	tty flush [-input] [-output] <fileId>
 * Flushes tty input and/or output buffers. Default is not to
 * flush.
 *==============================================================*/
int
tty_flush(ClientData cd, Tcl_Interp *interp, int argc, char *argv[])
{
  int		input = 0;
  int		output = 0;
  int		i, ret;
  FILE		*fp;

  struct option options[] = {
	{ "input",	T_TRUE,		(void*)&input		},
	{ "output",	T_TRUE,		(void*)&output		},
	{ NULL,		T_VOID,		NULL			}
  };


  /* Parse arguments */
  i = 1;
  if ((ret = tcl_getopt(interp, options, argc, argv, &i)) != TCL_OK) {
  	return TCL_ERROR;
  }
  if (i < argc-1) {
  	sprintf(interp->result, "%s: unknown option %s",
  					argv[0], argv[i]);
  	return TCL_ERROR;
  }

  /* Make sure we have a tty name */
  if (i != argc-1) {
  	sprintf(interp->result, "%s: need file_id argument", argv[0]);
  	return TCL_ERROR;
  }

  if (Tcl_GetOpenFile(interp, argv[1], 1, 0, &fp) != TCL_OK) {
  	sprintf(interp->result, "%s: no such file", argv[0]);
  	return TCL_ERROR;
  }
  mdm_flush(fileno(fp), input, output);
  return TCL_OK;
}

/*==============================================================*
 * tty_hangup		hangup a tty device
 * Usage:
 *	tty hangup <fileId>
 * Hangup the tty.
 *==============================================================*/
int
tty_hangup(ClientData cd, Tcl_Interp *interp, int argc, char *argv[])
{
  int		i, ret;
  FILE		*fp;

  struct option options[] = {
	{ NULL,		T_VOID,		NULL			}
  };


  /* Parse arguments */
  i = 1;
  if ((ret = tcl_getopt(interp, options, argc, argv, &i)) != TCL_OK) {
  	return TCL_ERROR;
  }
  if (i < argc-1) {
  	sprintf(interp->result, "%s: unknown option %s", argv[0], argv[i]);
  	return TCL_ERROR;
  }

  /* Make sure we have a tty name */
  if (i != argc-1) {
  	sprintf(interp->result, "%s: need file_id argument", argv[0]);
  	return TCL_ERROR;
  }

  if (Tcl_GetOpenFile(interp, argv[1], 1, 0, &fp) != TCL_OK) {
  	sprintf(interp->result, "%s: no such file", argv[0]);
  	return TCL_ERROR;
  }
  mdm_dtrtoggle(fileno(fp));
  return TCL_OK;
}

/*==============================================================*
 * Management functions for saving the tty state across TCL calls.
 * This code works for multiple ttys.
 *
 * tty_freestate	free an allocated state struct
 * tty_savestate	allocate a state struct, and save tty flags
 * tty_restorestate	restore saved tty state
 * tty_updatestate	set new port fd
 *==============================================================*/
static void
tty_freestate(int portfd)
{
  struct tty_state	*statp, *prevp = NULL;

  statp = tty_state;
  while (statp != NULL && statp->portfd != portfd) {
  	prevp = statp;
  	statp = statp->next;
  }
  if (statp != NULL) {
  	if (statp == tty_state) {
  		tty_state = statp->next;
  	} else {
  		prevp->next = statp->next;
  	}
  	free (statp->state);
  	free (statp);
  }
}

static int
tty_savestate(int portfd)
{
  struct tty_state	*statp;
  struct stat		stb;

  tty_freestate(portfd);
  if ((statp = malloc(sizeof(struct tty_state))) == NULL) {
  	return TCL_ERROR;
  }
  if ((statp->state = mdm_savestate(portfd)) == NULL) {
  	free(statp);
  	return TCL_ERROR;
  }
  if (fstat(portfd, &stb) != 0) {
  	free(statp);
  	return TCL_ERROR;
  }
  statp->portfd = portfd;
  statp->uid = stb.st_uid;
  statp->gid = stb.st_gid;
  statp->mode = stb.st_mode;
  statp->next = tty_state;
  if (ioctl(portfd, TIOCGETD, &statp->ldisc) < 0) {
  /*
   || ioctl(portfd, SIOCGIFENCAP, &statp->encap) < 0 ) {
  */
	free (statp->state);
  	free (statp);
  	return TCL_ERROR;
  }
  tty_state = statp;
  return TCL_OK;
}

static int
tty_restorestate(int portfd)
{
  struct tty_state	*statp;

  statp = tty_state;
  while (statp != NULL && statp->portfd != portfd) {
  	statp = statp->next;
  }
  if (statp == NULL) {
  	return TCL_ERROR;
  }
  /*
  ioctl(portfd, SIOCSIFENCAP, &statp->encap);
  */
  ioctl(portfd, TIOCSETD, &statp->ldisc);
  mdm_restorestate(portfd, statp->state);
  fchown(portfd, statp->uid, statp->gid);
  fchmod(portfd, statp->mode);
  tty_freestate(portfd);
  return TCL_OK;
}

static int
tty_updatestate(int portfd, int newfd)
{
  struct tty_state	*statp;

  statp = tty_state;
  while (statp != NULL && statp->portfd != portfd) {
  	statp = statp->next;
  }
  if (statp == NULL) {
  	return TCL_ERROR;
  }
  statp->portfd = newfd;
  return TCL_OK;
}

/*==============================================================*
 * Miscellaneous utility functions.
 *
 * basename	- return last component of a filename
 * devname	- return complete device file name
 * lockname	- return lock file name for given device file
 *==============================================================*/

#ifndef HAVE_BASENAME
/*
 * Return the last part of a filename.
 * Starting with Linux libc 4.6.something, a basename() function is
 * defined in unistd.h.
 */
static char *
basename(char *s)
{
  char *p;
  
  if((p = strrchr(s, '/')) == (char *)NULL)
  	p = s;
  else
  	p++;
  return(p);
}
#endif

/*
 * Construct name of device file
 */
static char *
devname(char *dev)
{
  static char	port[PATH_MAX];

  if (dev[0] == '/') {
  	strcpy( port, dev );
  } else {
  	sprintf( port, "/dev/%s", dev );
  }
  return port;
}

/*
 * Construct the lock file name
 */
static char *
lockname(char *lockdir, char *port)
{
  static char	lockfile[PATH_MAX];

  if (lockdir != NULL) {
  	sprintf(lockfile, "%s/LCK..%s", lockdir, basename(port));
  } else {
  	sprintf(lockfile, "%s/LCK..%s", TTY_LOCKDIR, basename(port));
  }
  return lockfile;
}

static int
tty_uid(char *user)
{
  struct passwd	*pwd;

  if ((pwd = getpwnam(user)) == NULL) {
  	return -1;
  }
  return pwd->pw_uid;
}

static int
tty_gid(char *group)
{
  struct group	*grp;

  if ((grp = getgrnam(group)) == NULL) {
  	return -1;
  }
  return grp->gr_gid;
}

